﻿import datetime
import json
import os
import os.path
import subprocess
import time

import pymel.core as pm
import MayaTest
import RunTests

# Important: these constants must match in C++ and Python (PerformanceTester.h / PerformanceTester.py)
TestRegistryDirectory = 'Ephere'
TestRegistryFormatVersion = 4
TestRegistryKeyVersion = 'version'
TestRegistryKeyModules = 'modules'
TestRegistryKeyTime = 'time'
TestRegistryKeyImprovementsHistory = 'improvementsHistory'
TestRegistryKeyLatestRuns = 'latestRuns'
TestRegistryKeyRecordDate = 'date'
TestRegistryKeyRecordRevision = 'revision'
TestRegistryKeyRecordChangeset = 'changeset'
TestRegistryKeyRecordTime = 'time'

PerformanceTestRepeats = 5
PerformanceTicksPerSecond = 1000000
FailTolerance = 0.08
MinimumConsecutiveFails = 5
MinimumConsecutiveImprovements = 5
ImproveTolerance = -0.02
TimeoutBetweenRuns = 0.2
LatestRunsCount = 10

# Set this to False while developing a new performance test to avoid storing any results in the registry
PerformanceTestRecordResults = True

# This class is no longer used, all performance testing is done by the C++ PerformanceTester in Ephere.NativeTools.Test, which now gets called by Python.
# We keep it in case we ever need to test performance of pure Python code (not based on C++) in the future
class PerformanceTester(object):

	def __init__( self, moduleName, isDebug ):
		self.moduleName = moduleName
		self.isDebug = isDebug

		homepath = os.path.expanduser( '~' )
		if homepath.endswith( '/Documents' ):
			homepath = homepath[:-9]

		self.testRegistryDir = os.path.join( homepath, TestRegistryDirectory )
		self.testRegistryFilePath = os.path.join( self.testRegistryDir, 'TestRegistry.json' )
		self.testRegistry = { TestRegistryKeyVersion: TestRegistryFormatVersion, TestRegistryKeyModules: {} }

		if not os.access( self.testRegistryDir, os.F_OK ):
			os.mkdir( self.testRegistryDir )
		if os.access( self.testRegistryFilePath, os.F_OK):
			with open( self.testRegistryFilePath ) as f:
				savedTestRegistry = json.load( f )
				if TestRegistryKeyVersion in savedTestRegistry \
						and savedTestRegistry[TestRegistryKeyVersion] == TestRegistryFormatVersion \
						and TestRegistryKeyModules in savedTestRegistry:
					self.testRegistry = savedTestRegistry
				else:
					# Backup the file that we found, it probably has an older format
					oldTestRegistryFilePath = os.path.join( self.testRegistryDir, 'TestRegistry{}.json'.format( datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S') ) )
					pm.mel.OxLog( 'Backing up old TestRegistry.json file to ' + oldTestRegistryFilePath )
					with open( oldTestRegistryFilePath, 'w' ) as f:
						json.dump( savedTestRegistry, f, sort_keys = True, indent = 4, separators = (',', ': ') )

		if not self.moduleName in self.testRegistry[TestRegistryKeyModules]:
			self.testRegistry[TestRegistryKeyModules][self.moduleName] = {}

		self.moduleRegistry = self.testRegistry[TestRegistryKeyModules][self.moduleName]

	def WritePerformanceTestResult( self, testName, time, isImprovement ):
		if os.getenv('PerformanceTestsRecordResults') == '0' or not PerformanceTestRecordResults:
			return

		currentTestRegistry = self.moduleRegistry[testName]

		latestRuns = currentTestRegistry[TestRegistryKeyLatestRuns]
		record = {
			TestRegistryKeyRecordDate : datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
			TestRegistryKeyRecordTime : time,
			TestRegistryKeyRecordRevision : RunTests.RepositoryRevision,
			TestRegistryKeyRecordChangeset : RunTests.RepositoryChangeset
		}

		latestRuns.append( record )
		if len( latestRuns ) > LatestRunsCount:
			del latestRuns[0]

		if isImprovement:
			currentTestRegistry[TestRegistryKeyTime] = time

			if not TestRegistryKeyImprovementsHistory in currentTestRegistry:
				currentTestRegistry[TestRegistryKeyImprovementsHistory] = []

			history = currentTestRegistry[TestRegistryKeyImprovementsHistory]
			history.append( record )
			if len( history ) > LatestRunsCount:
				del history[0]

		with open( self.testRegistryFilePath, 'w' ) as f:
			json.dump( self.testRegistry, f, sort_keys = True, indent = 4, separators = (',', ': ') )

	def Test( self, testName, setupCode, targetCode, printProfileLocation = None ):

		# Setting RunTests.PerformanceTestsMode = 0 will disable all performance tests
		if RunTests.PerformanceTestsMode <= 0:
			pm.mel.OxLog( 'Performance tests are globally disabled' )
			return

		if MayaTest.OsIsLinux or self.isDebug or RunTests.PerformanceTestsMode >= 2:
			if self.isDebug:
				pm.mel.OxLog( 'Performance tests are disabled in Debug configuration: "{}"'.format( testName ) )
				# Performance tests do not run in Debug configuration because they take too long. Comment-out this return to run your test once
				return

			if RunTests.PerformanceTestsMode >= 2:
				pm.mel.OxLog( 'Performance tests are set to "run once", running: "{}"'.format( testName ) )
			elif MayaTest.OsIsLinux:
				pm.mel.OxLog( 'Performance monitoring is disabled on Linux, running test for validation only: "{}"'.format( testName ) )

			if setupCode is not None:
				setupCode()

			totalTime = datetime.datetime.now()

			targetCode()

			totalTime = int( ( datetime.datetime.now() - totalTime ).total_seconds() * PerformanceTicksPerSecond )
			pm.mel.OxLog( 'Performance test "{}" ends: {}'.format( testName, totalTime ) )
			if setupCode is not None:
				pm.newFile( force = True )
			return

		previousBestTime = 0
		if testName in self.moduleRegistry:
			previousBestTime = self.moduleRegistry[testName][TestRegistryKeyTime] if TestRegistryKeyTime in self.moduleRegistry[testName] else 0
		else:
			self.moduleRegistry[testName] = { TestRegistryKeyImprovementsHistory : [] }

		bestTime = 0
		runIndex = 0
		actualRepeats = PerformanceTestRepeats if previousBestTime <= 0 else ( PerformanceTestRepeats + 2 )
		timeout = TimeoutBetweenRuns
		locationTotals = None
		failTime = previousBestTime * ( 1 + FailTolerance )
		while runIndex < actualRepeats and ( previousBestTime == 0 or bestTime == 0 or bestTime > failTime ):
			if runIndex > 0:
				time.sleep( timeout )
				if previousBestTime > 0 and timeout < 5:
					timeout = timeout * 2
			pm.mel.OxLog( 'Performance test "{}" run {} starts'.format( testName, runIndex ) )
			if setupCode is not None:
				setupCode()

			pm.mel.OxProfiler( '-start' )
			runTime = datetime.datetime.now()

			targetCode()

			runTime = int( ( datetime.datetime.now() - runTime ).total_seconds() * PerformanceTicksPerSecond )
			pm.mel.OxProfiler( '-stop' )
			pm.mel.OxLog( 'Performance test "{}" run {} ends: {}'.format( testName, runIndex, runTime ) )
			if setupCode is not None:
				pm.newFile( force = True )

			if runTime == 0:
				return 'Test {} is too fast or invalid'.format( testName )
			if bestTime == 0 or runTime < bestTime:
				bestTime = runTime
				if printProfileLocation is not None:
					locationTotals = pm.mel.OxProfiler( '-getLocation', printProfileLocation )
				pm.mel.OxProfiler( '-save', os.path.join( self.testRegistryDir, 'ProfilingResults_' + testName ) )

			pm.mel.OxProfiler( '-reset' )
			runIndex = runIndex + 1

		if not TestRegistryKeyLatestRuns in self.moduleRegistry[testName]:
			self.moduleRegistry[testName][TestRegistryKeyLatestRuns] = []

		pm.mel.OxLog( 'Performance test "{}" time: {} / {}'.format( testName, bestTime, previousBestTime ) )
		if printProfileLocation is not None:
			pm.mel.OxLog( 'Profiler location "{}" totals: calls = {}, exclusive = {}'.format( printProfileLocation, locationTotals[0], locationTotals[1] ) )
		isImprovement = False
		try:
			if previousBestTime > 0:
				changePercent = bestTime / float( previousBestTime ) - 1
				latestRuns = self.moduleRegistry[testName][TestRegistryKeyLatestRuns]
				if changePercent > FailTolerance:
					if MayaTest.OsIsLinux:
						subprocess.call("ps -Ao user,pid,pcpu,comm,cmd --sort=-pcpu --cols=120 | head", shell=True)
					elif MayaTest.OsIsMac:
						subprocess.call("ps -Ao user,pid,pcpu,comm -r | head", shell=True)

					if len( latestRuns ) >= MinimumConsecutiveFails - 1 \
						and all( map( lambda record: record[TestRegistryKeyRecordTime] > failTime, latestRuns[ 1-MinimumConsecutiveFails : ] ) ):
							# We have MinimumConsecutiveFails number of consective failed test runs, we can fail the test
							firstFailRevision = ''
							for i in range( -MinimumConsecutiveFails, -len( latestRuns ) - 1, -1 ):
								if latestRuns[i][TestRegistryKeyRecordTime] <= failTime:
									firstFailRevision = latestRuns[i+1][TestRegistryKeyRecordChangeset]
									break

							if firstFailRevision != '':
								firstFailRevision = " First failed changeset: {}".format( firstFailRevision )
							errorMessage = 'Test {} time: {} / {}, {:.0%} performance degradation!{}'.format( testName, bestTime, previousBestTime, changePercent, firstFailRevision );
							pm.mel.OxLog( errorMessage, '-l', 'warning' )
							return errorMessage
					else:
						pm.mel.OxLog( 'Test {} time: {} / {}, performance may be degrading! {:.0%} decrease'
								.format( testName, bestTime, previousBestTime, changePercent ), '-l', 'warning' )
						if os.getenv( 'TriggeredBy' ) != '':
							print( "##teamcity[setParameter name='env.TagsToAdd' value='{} PerformanceWarning']".format( os.getenv( 'TagsToAdd' ) or '' ) )
				else:
					improveTime = previousBestTime * ( 1 + ImproveTolerance )

					if bestTime < improveTime \
						and len( latestRuns ) >= MinimumConsecutiveImprovements - 1 \
						and all( map( lambda record: record[TestRegistryKeyRecordTime] < improveTime, latestRuns[ 1-MinimumConsecutiveImprovements : ] ) ):
						isImprovement = True

					pm.mel.OxLog( 'Test {} time: {} / {}{}, difference: {:.0%}'
						.format( testName, bestTime, previousBestTime, ', improved result recorded' if isImprovement else '', changePercent ) )
					if os.getenv( 'TriggeredBy' ) != '' and isImprovement:
						print( "##teamcity[setParameter name='env.TagsToAdd' value='{} PerformanceImproved']".format( os.getenv( 'TagsToAdd' ) or '' ) )
			else:
				isImprovement = True
				pm.mel.OxLog( 'Test {} initial time recorded: {}'.format( testName, bestTime ) )
		finally:
			self.WritePerformanceTestResult( testName, bestTime, isImprovement )

# Not very reliable
def MeasureMemoryUsage( code ):
	result = 0
	for i in range( 0, 5 ):
		memoryBefore = pm.mel.OxProfiler( '-mu' )
		code()
		memoryDelta = pm.mel.OxProfiler( '-mu' ) - memoryBefore
		if result == 0 or result > memoryDelta:
			result = memoryDelta
		pm.newFile( force = True )

	return result
