﻿import MayaTest
import TestUtilities
import pymel.core as pm
import pymel.core.datatypes as dt
import os

TestsPath = os.path.dirname( __file__ ) if '__file__' in globals() and len( __file__ ) > 0 else os.getcwd()

class Test_DynamicsNode( MayaTest.OxTestCase ):

	@MayaTest.SkipIfNoEngines
	def test_BasicDynamics( self ):
		guidesShape = TestUtilities.AddGuidesToNewPlane()
		guidesFromMesh = TestUtilities.GetNodeByType( TestUtilities.GuidesFromMeshNodeName )
		dynamics = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.DynamicsNodeName ) )

		verticesAtFrame1 = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		pm.currentTime( 2 )
		pm.currentTime( 3 )
		verticesAtFrame2 = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		# The vertices should have moved at frame 3
		TestUtilities.CheckPointsNotAllNearEqual( self, verticesAtFrame1, verticesAtFrame2 )

	def SetupDeformingMeshWithHairDynamics( self, startTime = 0, endTime = 5, animatedVertexOffset = 6 ):
		plane = pm.polyPlane( sx = 1, sy = 1, h = 10, w = 10 )
		meshShape = plane[0]

		# Animate plane vertices
		pm.select( meshShape + ".vtx[0] ", replace = True )
		pm.select( meshShape + ".vtx[2]", add = True  )

		pm.currentTime( startTime )
		pm.setKeyframe()

		pm.currentTime( endTime )
		pm.move( -animatedVertexOffset, 0, 0, r = True )
		pm.setKeyframe()

		pm.select( meshShape + ".vtx[1] ", replace = True )
		pm.select( meshShape + ".vtx[3]", add = True  )

		pm.currentTime( startTime )
		pm.setKeyframe()

		pm.currentTime( endTime )
		pm.move( animatedVertexOffset, 0, 0, r = True )
		pm.setKeyframe()

		pm.select( plane, replace = True )
		guidesShape = TestUtilities.AddGuidesToMesh( meshShape, 4, 4, 2, 0, 10 )

		guidesFromMesh = TestUtilities.GetNodeByType( TestUtilities.GuidesFromMeshNodeName )
		dynamicsName = pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.DynamicsNodeName )
		dynamics = pm.PyNode( dynamicsName )

		# Make gravity go up
		dynamics.zGravity.set( 998.00 )
		dynamics.yGravity.set( 0 )
		dynamics.baseMeshAnimType.set( 2 ) # set base mesh deformable

		return guidesShape, meshShape, dynamics

	def test_DeformingBaseMesh( self ):
		# This test currently works even without any physics adapters
		guidesAndMesh = self.SetupDeformingMeshWithHairDynamics()
		guidesShape = guidesAndMesh[0]
		meshShape = guidesAndMesh[1]

		# At all frames the roots of the strands must be in the same positions as mesh vertices
		for frame in range( 0, 10 ):
			pm.currentTime( frame )
			for vertexIndex in range( 0, 4 ):
				meshVertices = TestUtilities.GetMeshVertices( meshShape )
				rootPosition = TestUtilities.GetStrandPointPositionsAtIndex( guidesShape, 0, True )
				TestUtilities.CheckPointsAllNearEqual( self, meshVertices, rootPosition )

	@MayaTest.SkipIfNoEngines
	def test_RecordDynamics( self ):

		guidesShape = TestUtilities.AddGuidesToNewPlane()

		guidesFromMesh = TestUtilities.GetNodeByType( TestUtilities.GuidesFromMeshNodeName )
		dynamicsName = pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.DynamicsNodeName )
		dynamics = pm.PyNode( dynamicsName )

		# Change the gravity vector to go sideways so we get more motion from the hair
		dynamics.zGravity.set( 0 )
		dynamics.yGravity.set( -998.000 )

		# Record dynamics for current time range
		pm.playbackOptions( animationStartTime = 1 )
		pm.playbackOptions( animationEndTime = 5 )

		# This forces initial call to compute (to initialize the simulator)
		pm.currentTime( 1 )
		previousVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		pm.mel.OxRecordDynamics( dynamics )

		# Going forward or backward, each frame should have different vertices
		pm.currentTime( 1 )
		previousVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		for currentFrame in range( 2, 6 ):
			pm.currentTime( currentFrame )
			currentVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
			TestUtilities.CheckPointsNotAllNearEqual( self, previousVertices, currentVertices )
			previousVertices = currentVertices

		for currentFrame in range( 4, 0, -1 ):
			pm.currentTime( currentFrame )
			currentVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
			TestUtilities.CheckPointsNotAllNearEqual( self, previousVertices, currentVertices )
			previousVertices = currentVertices

		pm.currentTime( 3 )
		frame3Vertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		# Test save/load scene with recorded dynamics
		filePath = pm.saveAs( 'temp.mb' )
		pm.newFile( force = True )
		pm.openFile( filePath, force = True )
		guidesShape = TestUtilities.GetNodeByType( TestUtilities.HairShapeName )
		pm.currentTime( 3 )
		currentVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		TestUtilities.CheckPointsAllNearEqual( self, frame3Vertices, currentVertices )

		filePath = pm.saveAs( 'temp.ma' )
		pm.newFile( force = True )
		pm.openFile( filePath, force = True )
		guidesShape = TestUtilities.GetNodeByType( TestUtilities.HairShapeName )
		pm.currentTime( 3 )
		currentVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		TestUtilities.CheckPointsAllNearEqual( self, frame3Vertices, currentVertices )

	def test_RecordedPlaybackOnDeformingMesh( self ):
		# This test currently works even without any physics adapters

		startTime = 0
		endTime = 5

		guidesAndMesh = self.SetupDeformingMeshWithHairDynamics( startTime, endTime, 20 )
		guidesShape = guidesAndMesh[0]
		meshShape = guidesAndMesh[1]
		dynamics = guidesAndMesh[2]

		# Record frames 0-5
		pm.playbackOptions( animationStartTime = startTime )
		pm.playbackOptions( animationEndTime = endTime )
		pm.mel.OxRecordDynamics( dynamics )

		# At all frames the roots of the strands must be in the same positions as mesh vertices
		pm.currentTime( startTime )
		for frame in range( startTime, endTime + 1 ):
			pm.currentTime( frame )
			for vertexIndex in range( 0, 4 ):
				meshVertices = TestUtilities.GetMeshVertices( meshShape )
				rootPosition = TestUtilities.GetStrandPointPositionsAtIndex( guidesShape, 0, True )
				TestUtilities.CheckPointsAllNearEqual( self, meshVertices, rootPosition )

	def test_LoadRecordedDynamics( self ):
		# This test must work even without any physics adapters
		# It also tests compatibility with older scenes that use GuidesShape. A more extensive separate test of that can be added if needed

		pm.openFile( TestsPath + '/RecordedDynamics.mb', force = True )
		guidesShape = TestUtilities.GetNodeByType( TestUtilities.GuidesShapeName )
		pm.currentTime( 3 )
		currentVertices = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		frame3Vertices = [[-0.5, 0.0, 0.5], [-3.4990067481994678, 49.439109802246094, 3.4989922046661275], [0.5, 0.0, 0.5],
			[3.497434139251704, 49.4390754699707, 3.4974119663238405], [-0.5, 0.0, -0.5], [-3.4947767257690487, 49.43912887573242, -3.494731664657603],
			[0.5, -2.465190328815662e-32, -0.5], [3.490152706966748, 49.43867492675783, -3.4902481274381683]]
		TestUtilities.CheckPointsAllNearEqual( self, frame3Vertices, currentVertices )

	@MayaTest.SkipIfNoEngines
	def test_ExternalForce( self ):

		airFieldNode = pm.mel.air(magnitude = 8.0)[0]

		guidesShape = TestUtilities.AddGuidesToNewPlane()
		guidesFromMesh = TestUtilities.GetNodeByType( TestUtilities.GuidesFromMeshNodeName )
		guidesFromMesh.length.set( 4 )
		dynamicsName = pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.DynamicsNodeName )
		dynamics = pm.PyNode( dynamicsName )

		# No internal forces
		dynamics.xGravity.set( 0 )
		dynamics.yGravity.set( 0 )
		dynamics.zGravity.set( 0 )

		pm.mel.source( "AEDynamicsNodeTemplate.mel" )

		pm.mel.OxConnectForceField( airFieldNode, dynamics.forceFields )

		pm.currentTime( 1 )
		verticesAtFrame1 = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )
		pm.currentTime( 2 )
		verticesAtFrame2 = TestUtilities.GetVerticesInObjectCoordinates( guidesShape )

		# The vertices should have moved at frame 2 but not by too much
		averagePointDistance = TestUtilities.GetAveragePointDistance( self, verticesAtFrame1, verticesAtFrame2 )
		self.assertLess( averagePointDistance, 1.0 )
		self.assertGreater( averagePointDistance, 0.1 )

	def SetupComputePerformance( self ):
		self.topShape = TestUtilities.AddGuidesToNewPlane( guideCount = 300, pointsPerStrandCount = 10, rootGenerationMethod = 2, planeSegmentCount = 2, planeSize = 100 )
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]
		dynamics = pm.PyNode( pm.mel.OxAddStrandOperator( guidesFromMesh, TestUtilities.DynamicsNodeName ) )

		pm.currentTime( 1 )
		if MayaTest.InBatchMode:
			self.topShape.boundingBox()
		else:
			pm.refresh()

	@MayaTest.SkipIfNoEngines
	def test_ComputePerformance( self ):
		self.assertPerformance( 'MayaTest.Self.SetupComputePerformance()', 'MayaTest.RunAnimation( MayaTest.Self.topShape, startFrame = 2, frameCount = 99 )' )
