﻿import MayaTest
import TestUtilities
import os.path
import pymel.core as pm
import pymel.core.datatypes as dt
import xml.etree.ElementTree as et
import unittest

''' This class tests that all of the operators within Ornatrix Maya can be properly saved and loaded as grooms. It needs to be updated whenever new operators which support
grooms are added, or when new groom parameters are added '''
class Test_Grooms( MayaTest.OxTestCase ):

	# Tests for a bug where default interpolation parameter is specified in HairFromGuides groom operator causing an assertion when generating hair
	def test_BarycentricAttributePresentInHFGGroom( self ):
		hairShape = pm.mel.OxLoadGroom( path = """<?xml version="1.0" encoding="utf-8"?>
<MayaGroom name="test_BarycentricAttributePresentInHFGGroom">
	<Stack>
		<Sphere referenced="true" optional="true" />
		<GuidesFromMesh />
		<HairFromGuides interpolation="Barycentric" />
	</Stack>
</MayaGroom>""" )

		strandCount = pm.mel.OxGetStrandCount( hairShape )
		self.assertTrue( strandCount != 0 )

	def TestGroom( self, operatorNodeName, operatorName, nonDefaultParameters, expectedNonDefaultAttributes = [], addGuides = True, addHair = False ):
		defaultParametersGroomName = "test_groomDefaultParameters" + operatorName
		nonDefaultParametersGroomName = "test_groomNonDefaultParameters" + operatorName

		guidesShape = None
		expectedOperators = None
		operator = None
		if addGuides:
			if addHair:
				guidesShape = TestUtilities.AddHairToNewPlane()
				expectedOperators = ["PolygonMesh", "GuidesFromMesh", "HairFromGuides", operatorName]
			else:
				guidesShape = TestUtilities.AddGuidesToNewPlane()
				expectedOperators = ["PolygonMesh", "GuidesFromMesh", operatorName]
			operator = pm.PyNode( pm.mel.OxAddStrandOperator( guidesShape, operatorNodeName ) )
		else:
			expectedOperators = ["PolygonMesh", operatorName]
			if operatorNodeName == TestUtilities.GuidesFromMeshNodeName:
				plane = pm.polyPlane()
				pm.select( plane )
				meshShape = plane[0]

				# Create nodes
				transformNode = meshShape.firstParent2()
				operator = pm.createNode( TestUtilities.GuidesFromMeshNodeName )
				guidesShape = pm.createNode( TestUtilities.HairShapeName, parent = transformNode )

				# Make connections
				meshShape.outMesh >> operator.inputMesh
				operator.outputGuides >> guidesShape.inputHair
				# Note: `guidesShape.distributionMesh` isn't connected here
			elif operatorNodeName == TestUtilities.GuidesFromCurvesNodeName:
				curve1Root = (1, 0, 1)
				curve1Offsets = (0, 3, 0.3)
				curve1 = pm.curve( p=[curve1Root, (1, 1, 1.1), (1, 2, 1.2), ( curve1Root[0] + curve1Offsets[0], curve1Root[1] + curve1Offsets[1], curve1Root[2] + curve1Offsets[2] )], k=[0, 0, 0, 1, 1, 1] )
				curve2 = pm.curve( p=[(-1, 0, 1), (-1, 1, 1.1), (-1, 2, 1.2), (-1, 3, 1.3)], k=[0, 0, 0, 1, 1, 1] )

				pm.select( curve1, replace = True )
				pm.select( curve2, add = True )
				guidesShape = pm.PyNode( pm.mel.OxAddGuidesFromCurves() )
				operator = pm.ls( type = TestUtilities.GuidesFromCurvesNodeName )[0]
				expectedOperators = ["Curves", operatorName]
			else:
				plane = pm.polyPlane( sx = 1, sy = 1 )
				pm.select( plane )
				guidesShape = plane[0]

		pm.mel.OxSaveGroom( path = self.addTempFile( defaultParametersGroomName ), optional = True )
		defaultParametersGroomFilePath = self.addTempFile( defaultParametersGroomName + ".oxmaya" )
		xml = et.parse( defaultParametersGroomFilePath )
		root = xml.getroot()[0]
		children = [child.tag for child in root]
		self.assertEqual( children, expectedOperators )
		operatorNode = root.find( operatorName )

		#print( defaultParametersGroomFilePath )

		# With default values we have an empty operator element
		operatorChildren = [child.tag for child in operatorNode]
		self.assertEqual( [], operatorChildren )

		# Uncomment to see te XML file
		#with open( defaultParametersGroomFilePath, 'r' ) as fin:
		#	print fin.read()

		operatorAttributes = [attrib
					#+ ': ' + str( operatorNode.get( attrib ) )
					for attrib in operatorNode.attrib]
		self.assertEqual( expectedNonDefaultAttributes, operatorAttributes )

		# Set non-default values
		for parameter in nonDefaultParameters:
			mayaAttribute = operator.attr( parameter[0] )

			attributeValue = 0
			if len( parameter ) > 1:
				attributeValue = parameter[1]
			else:
				# Values which typically would never be used as a default
				nonDefaultValues = {
					'float' : 0.234,
					'long' : 87,
					'integer' : 87,
					'enum' : 12,
					'string' : 'foo' }
				attributeValue = nonDefaultValues[mayaAttribute.type()]

			mayaAttribute.set( attributeValue )

		pm.mel.OxSaveGroom( path = self.addTempFile( nonDefaultParametersGroomName ), optional = True )
		xml = et.parse( self.addTempFile( nonDefaultParametersGroomName + ".oxmaya" ) )

		root = xml.getroot()[0]
		children = [child.tag for child in root]
		self.assertEqual( children, expectedOperators )
		operatorNode = root.find( operatorName )

		# With default values we have an empty operator element
		operatorChildren = [child.tag for child in operatorNode]
		operatorAttributes = [attrib for attrib in operatorNode.attrib]
		self.assertEqual( len( nonDefaultParameters ) + len( expectedNonDefaultAttributes ), len( operatorAttributes ) + len( operatorChildren ), str( operatorAttributes ) + str( operatorChildren ) )

	def test_CurlGroom( self ):
		self.TestGroom( TestUtilities.CurlNodeName, "Curl", [
				["waveform", 0],
				["magnitude"],
				["phase"],
				["phaseOffset"],
				["stretch"],
				["isLengthDependent", True],
				["noiseScale"],
				["noiseAmount"]],
			['magnitudeCurve'] )

	def test_DetailGroom( self ):
		self.TestGroom( TestUtilities.DetailNodeName, "Detail", [
				["viewPointCount"],
				["renderPointCount"],
				["angleThreshold"],
				["sampleByThreshold", True],
				["sampleSize"],
				["smoothing"]] )

	def test_DynamicsGroom( self ):
		self.TestGroom( TestUtilities.DynamicsNodeName, "HairDynamics", [
				["strandGroup"],
				["randomSeed"],
				["collideWithSurface", True],
				["surfaceSdfResolution"],
				["damping"],
				["particleRadius"],
				["subStepCount"],
				["iterationCount"],
				["stiffness"],
				["spacingOffset"],
				["maximumVelocity", 123.0],
				["latticeCount"],
				["latticeSize"],
				["groupStiffness"],
				["groupHolderGridPosition"]] )

	def test_EditGuidesGroom( self ):
		self.TestGroom( TestUtilities.EditGuidesShapeName, "EditGuides", [
				["applyAmount"]],
				["applyRamp"])

	def test_FrizzGroom( self ):
		self.TestGroom( TestUtilities.FrizzNodeName, "Frizz", [
				["amount"],
				["scale"],
				["outlierFraction"],
				["outlierAmount"],
				["randomSeed"],
				["groupId"] ],
			['amountCurve'] )

	def test_GenerateStrandData( self ):
		self.TestGroom( TestUtilities.GenerateStrandDataNodeName, "GenerateStrandData", [
				["strandGroup"],
				["randomSeed"],
				["newChannelType", 1],
				["newChannelName"],
				["generationMethod", 2],
				["useTargetValueRange", False],
				["minimumTargetValue"],
				["maximumTargetValue"],
				["useSourceValueRange", True],
				["minimumSourceValue"],
				["maximumSourceValue"]] )

	def test_GravityGroom( self ):
		self.TestGroom( TestUtilities.GravityNodeName, "Gravity", [
				["force"],
				["repel"],
				["strandGroup"]],
				["amountCurve"] )

	def test_GroundStrandsGroom( self ):
		self.TestGroom( TestUtilities.GroundStrandsNodeName, "GroundStrands", [] )

	def test_GuideClusterGroom( self ):
		self.TestGroom( TestUtilities.GuideClusterNodeName, "GuideCluster", [
				["clusterAmount"],
				["twistAmount"],
				["wispAmount"],
				["wispOffset"],
				["isUsingWisps", True]],
				["twistCurve", "clusterCurve"] )

	def test_GuidesFromCurvesGroom( self ):
		self.TestGroom( TestUtilities.GuidesFromCurvesNodeName, "GuidesFromCurves", [["pointCount"]], addGuides = False )

	def test_GuidesFromHairGroom( self ):
		self.TestGroom( TestUtilities.GuidesFromHairNodeName, "GuidesFromHair", [], addHair = True )

	def test_GuidesFromMeshGroom( self ):
		self.TestGroom( TestUtilities.GuidesFromMeshNodeName, "GuidesFromMesh", [
				["distribution", 4],
				["count"],
				["length"],
				["lengthRandomness"],
				["pointCount"],
				["randomSeed"],
				["topologyBasedOrientation", False],
				["detachRoots", True],
				["rememberRootPositions", True]], addGuides = False, addHair = False )

	def test_HairClusterGroom( self ):
		self.TestGroom( TestUtilities.HairClusterNodeName, "HairCluster", [
				["clusterCount"],
				["randomSeed"],
				["distanceThreshold"],
				["useUv", True]],
				["clusterCurve"], addHair = True )

	def test_HairFromGuidesGroom( self ):
		self.TestGroom( TestUtilities.HairFromGuidesNodeName, "HairFromGuides", [
				["distribution", 4],
				["viewportCount"],
				["renderCount"],
				["interpolationGuideCount", 1],
				["topologyBasedOrientation", False],
				["useGuideProximity", True],
				["guideProximityDistance"]],
				["interpolation"] )

	def test_LengthGroom( self ):
		self.TestGroom( TestUtilities.LengthNodeName, "Length", [
				["value"],
				["randomize"],
				["randomSeed"],
				["strandGroup"],
				["minimumValue"],
				["maximumValue"]] )

	def test_MeshFromHairGroom( self ):
		self.TestGroom( TestUtilities.MeshFromStrandsNodeName, "MeshFromStrands", [
				["meshType", 0],
				["strandGroup"],
				["randomSeed"]] )

	def test_MultiplierGroom( self ):
		self.TestGroom( TestUtilities.MultiplierNodeName, "Multiplier", [
				["randomSeed"],
				["strandGroup"],
				["probability"],
				["copies"],
				["copiesViewPercent"],
				["spreadRoot"],
				["spreadTip"],
				["fluffRoot"],
				["fluffTip"],
				["shiftRoot"],
				["shiftTip"],
				["twist"]],
				["twistCurve", "spreadCurve", "fluffCurve", "shiftCurve"] )

	def test_PropagationGroom( self ):
		self.TestGroom( TestUtilities.PropagationNodeName, "Propagation", [
				["generationMethod", 0],
				["count", 1],
				["verticesPerRootCount"],
				["strandGroup"],
				["countPerStrand", False],
				["randomSeed"],
				["lowRange"],
				["highRange"],
				["lengthRelativeToBaseStrand", True],
				["faceSurface", False],
				["sideCount"],
				["twist"],
				["twistRandomness"],
				["fanning"],
				["fanningRandomness"],
				["length"],
				["lengthRandomness"],
				["resultStrandGroup"]],
				["twistCurve", "fanningCurve", "lengthCurve", "distributionCurve"] )

	def test_PushAwayFromSurfaceGroom( self ):
		self.TestGroom( TestUtilities.PushAwayFromSurfaceNodeName, "PushAwayFromSurface", [
				["strandGroup"],
				["distance"]] )

	def test_RenderSettingsGroom( self ):
		self.TestGroom( TestUtilities.RenderSettingsNodeName, "RenderSettings", [
				["radius"],
				["strandGroup"]],
				["radiusCurve"] )

	def test_RotateGroom( self ):
		self.TestGroom( TestUtilities.RotateNodeName, "RotateStrands", [
				["globalAngle"],
				["globalChaos"],
				["faceCoordinate", True],
				["randomSeed"],
				["groupId"],
				["orientBasedOnStrandShape", True]] )

	def test_SurfaceCombGroom( self ):
		self.TestGroom( TestUtilities.SurfaceCombNodeName, "SurfaceComb", [
				["affectWholeStrand", True],
				["chaos"],
				["randomSeed"],
				["strandGroup"],
				["applyToShape", False],
				["applyToDirection", False]],
				["slopeCurve"] )

	def test_SymmetryGroom( self ):
		self.TestGroom( TestUtilities.SymmetryNodeName, "Symmetry", [
				["useIgnoreDistance", True],
				["ignoreDistance"]] )

	def test_GroomWithSphereAsBaseObject( self ):
		groomPath = self.addTempFile( "test_GroomWithSphereAsBaseObject.oxmaya" )
		hairShape = TestUtilities.AddHairToNewSphere( radius = 12.3 )
		pm.select( hairShape )
		pm.mel.OxSaveGroom( path = groomPath, optional = True )

		xml = et.parse( groomPath )
		root = xml.getroot()[0]
		children = [child.tag for child in root]
		self.assertEqual( children, ["Sphere", "GuidesFromMesh", "HairFromGuides"] )
		self.assertEqual( '12.3', root.find( 'Sphere' ).attrib["radius"] )
		self.assertEqual( 'true', root.find( 'Sphere' ).attrib["optional"] )

		# Load groom with no selection
		pm.newFile( force = 1 )
		hairShape = pm.mel.OxLoadGroom( path = groomPath )

		# Ensure that a sphere has been created as a base object and radius is correct
		sphereShape = pm.ls( type = "polySphere" )[0]
		self.assertAlmostEqual( 12.3, sphereShape.radius.get(), places = 3 )

	def test_CreateAndResaveFurballGroom( self ):
		hairShape = pm.PyNode( pm.mel.OxLoadGroom( path = "Fur Ball" ) )
		verticesFromStockGroom = TestUtilities.GetVerticesInObjectCoordinates( hairShape )
		verticesFromStockGroom.sort()

		# Generate second groom
		groomPath = self.addTempFile( "test_CreateAndResaveFurballGroom.oxmaya" )

		pm.mel.OxSaveGroom( path = groomPath, optional = True )
		pm.newFile( force = 1 )
		hairShape = pm.PyNode( pm.mel.OxLoadGroom( path = groomPath ) )
		verticesFromResavedGroom = TestUtilities.GetVerticesInObjectCoordinates( hairShape )
		verticesFromResavedGroom.sort()

		TestUtilities.CheckPointsAllNearEqual( self, verticesFromStockGroom, verticesFromResavedGroom )

	def test_FeathersGroom( self ):
		meshShape = pm.PyNode( pm.mel.OxLoadGroom( path = "Feathers" ) )
		verticesFromStockGroom = TestUtilities.GetMeshVertices( meshShape )
		self.assertGreater( len( verticesFromStockGroom ), 0 )

	def test_GroomCommands( self ):
		# Save grooms
		hairShape = TestUtilities.AddHairToNewPlane( rootGenerationMethod = 0 )
		hairFromGuides = TestUtilities.GetNodeByType( TestUtilities.HairFromGuidesNodeName )
		hairFromGuides.renderCount.set( 5000 )
		strandCount = pm.mel.OxGetStrandCount( hairShape )
		self.assertTrue( strandCount > 0 )
		pm.select( hairShape )
		pm.mel.OxSaveGroom( path = self.addTempFile( "test_groom" ), optional = True )

		xml = et.parse( self.addTempFile( "test_groom.oxmaya" ) )
		root = xml.getroot()[0]
		children = [child.tag for child in root]
		self.assertEqual( children, ["PolygonMesh", "GuidesFromMesh", "HairFromGuides"] )
		self.assertEqual( '5000', root.find( 'HairFromGuides' ).attrib["renderCount"] )

		# Load groom with no selection
		pm.newFile( force = 1 )
		with self.assertRaises( pm.MelError ):
			pm.mel.OxLoadGroom( path = "Invalid Groom Name" )
		hairShape = pm.mel.OxLoadGroom( path = self.addTempFile( "test_groom.oxmaya" ) )
		self.assertEqual( hairShape, "test_groomShape1" )
		hairFromGuides = TestUtilities.GetNodeByType( TestUtilities.HairFromGuidesNodeName )
		self.assertTrue( hairFromGuides.exists() )
		self.assertEqual( hairFromGuides.distribution.get(), 0 )
		self.assertAlmostEqual( hairFromGuides.viewportCountFraction.get(), 0.1 )
		# TODO: self.assertEqual( pm.mel.OxGetStrandCount( hairShape ), strandCount )

		# Load groom to selection
		pm.newFile( force = 1 )
		pm.polyPlane( ch = True, o = True, w = 10, h = 10, sh = 1, sw = 1 )
		pm.select( "pPlane1" )
		pm.flushUndo()

		pm.mel.OxLoadGroom( path = self.addTempFile( "test_groom.oxmaya" ) )
		hairFromGuides = TestUtilities.GetNodeByType( TestUtilities.HairFromGuidesNodeName )
		self.assertTrue( hairFromGuides.exists() )
		self.assertEqual( hairFromGuides.distribution.get(), 0 )
		self.assertAlmostEqual( hairFromGuides.viewportCountFraction.get(), 0.1 )
		hairShape = TestUtilities.GetNodeByType( TestUtilities.HairShapeName )
		self.assertNotEqual( len( hairShape.instObjGroups[0].outputs() ), 0 )

		# Undo
		pm.undo()
		self.assertEqual( [], pm.ls( type = TestUtilities.HairFromGuidesNodeName ) )

		# Redo
		pm.redo()
		hairFromGuides = TestUtilities.GetNodeByType( TestUtilities.HairFromGuidesNodeName )
		self.assertTrue( hairFromGuides.exists() )
		self.assertEqual( hairFromGuides.distribution.get(), 0 )
		self.assertAlmostEqual( hairFromGuides.viewportCountFraction.get(), 0.1 )

	def test_GroomSaveLoadEditGuides( self ):
		filePath = self.addTempFile( "test_groom.oxmaya" )
		plane = pm.polyPlane( sx = 1, sy = 1 )[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( plane, 2, 2 )
		guidesFromMesh = pm.ls( type = TestUtilities.GuidesFromMeshNodeName )[0]

		pm.select( editGuidesShape )
		pm.mel.OxEditGuides( 2, 0, 2, 0, 0, 0, 4, 0, 0, 1, 2, 0, 0, 0, 4, 0, 0, eg = True )
		originalPosition = pm.mel.OxGetStrandPoint( editGuidesShape, 0, 1 )
		pm.mel.OxSaveGroom( path = filePath, optional = True )
		pm.newFile( force = 1 )

		pm.polyPlane( sx = 1, sy = 1 )
		guidesShape = pm.PyNode( pm.mel.OxLoadGroom( path = filePath ) )
		editGuidesShape = TestUtilities.GetNodeByType( TestUtilities.EditGuidesShapeName )
		self.assertTrue( guidesShape.exists() )
		self.assertTrue( editGuidesShape.exists() )
		groomPositions = [pm.mel.OxGetStrandPoint( editGuidesShape, 0, 1 ), pm.mel.OxGetStrandPoint( editGuidesShape, 1, 1 )]

		hasOriginalPosition = False
		for groomPoint in groomPositions:
			if originalPosition == groomPoint:
				hasOriginalPosition = True

		self.assertTrue( hasOriginalPosition )

	def test_GroomSaveLoadBakedGuides( self ):
		plane = pm.polyPlane( sx = 1, sy = 1 )[0]
		editGuidesShape = TestUtilities.AddEditGuidesToMesh( plane, 2, 2 )
		guidesShape = pm.ls( type = TestUtilities.HairShapeName )[0]
		pm.select( editGuidesShape )
		pm.mel.eval( 'OxEditGuides -eg 2 0 2 0 0 0 4 0 0 1 2 0 0 0 4 0 0' )
		originalPosition = pm.mel.OxGetStrandPoint( guidesShape, 0, 1 )
		bakedGuides = pm.PyNode( pm.mel.OxCollapseStack( editGuidesShape ) )
		pm.mel.OxSaveGroom( path = self.addTempFile( "test_groom.oxmaya" ), optional = True )
		pm.newFile( force = 1 )

		guidesShape = pm.PyNode( pm.mel.OxLoadGroom( path = self.addTempFile( "test_groom.oxmaya" ) ) )
		self.assertTrue( guidesShape.exists() )
		self.assertEqual( guidesShape.type(), TestUtilities.HairShapeName )
		self.assertEqual( guidesShape.inputHair.inputs()[0].type(), TestUtilities.GuidesFromCurvesNodeName )
		#self.assertEqual( len( pm.ls( type = TestUtilities.GuidesFromCurvesNodeName ) ), 1 )
		self.assertEqual( len( pm.ls( type = 'nurbsCurve' ) ), 2 )
		self.assertEqual( pm.mel.OxGetStrandCount( guidesShape ), 2 )
		groomPosition = pm.mel.OxGetStrandPoint( guidesShape, 0, 1 )
		# TODO: figure out why points are in different coordinate systems and x/y/z are in different order
		#self.assertAlmostEqual( originalPosition[0], groomPosition[0], places = 1 )
