from .AutoLayout import *

from .Node.NodeDB.NodeFactory import NodeFactory
from .Node.NodeDB.PackedNetworkDummieNode import *
from .Node.ConnectionBuilder import *
from .Node.Connector import Connector
from .NodeCreationMenu import *


# import maya.cmds as mc


class OxScene(QtWidgets.QGraphicsScene):
	sceneUpdated = QtCore.Signal()
	newSelection = QtCore.Signal(object)
	grabStarted = QtCore.Signal(object)
	grabFinished = QtCore.Signal()

	def __init__(self, container):
		super(OxScene, self).__init__()
		self.container = container


		self.nodeList = {}
		self.dummiesNodeInput = None
		self.dummiesNodeOutput = None

		self.poolId = 0
		self.setSceneRect(-1000, -1000, 2000, 2000)
		self.AL = True
		self.totalTime = 0.0001
		self.hairSystem = None
		self.parent = None
		self.stackModel = None

		self.graphPlugin = {}

		# self.changed.connect(self.test)

		self.connectionBuilder = ConnectionBuilder(self)

		self.container.registerUI("connectionManager", self.connectionBuilder)
		self.container.registerUI("graph", self)

		self.deferedCall = None

		self.update()

		self.beginClick = False
		self.haveMoved = False
		self.ignoreNextCompute = False

	###
	### Graph Initialisation
	###


	###
	### Operation
	###

	def update(self):
		hairSystems = self.container.bridge.getHairSystem()
		if len(hairSystems):
			hsid = self.container.currentHairSystem
			self.setHairSystem(hairSystems[hsid])
			self.flushGraph()
			if self.hairSystem:
				self.container.bridge.saveHairNetwork(self.hairSystem)
				self.buildGraph()
				self.calculateTotalTime()
			self.sceneUpdated.emit()

	def rebuildConnection(self):
		for k, node in self.nodeList.items():
			if node.nodeContainer.isHostNode():
				for input in range(0, len(node.nodeContainer.getHostNode().getInputs())):
					sourceNodes = node.nodeContainer.getHostNode().getNodeUnder(input)
					for sourceNode in sourceNodes:
						if sourceNode.uid in self.nodeList:
							targetNode = self.nodeList[sourceNode.uid]
							for outputId, outConn in sourceNode.outputConn.iteritems():
								if outConn[0] == node.nodeContainer.getHostNode():
									if node.pinManager.getInputPin(
											input) is not None and targetNode.pinManager.getOutputPin(outputId) is not None:
										inPin = node.pinManager.getInputPin(input)
										outPin = targetNode.pinManager.getOutputPin(outputId)
										newConn = Connector(outPin, inPin, self.container)
										newConn.doConnect(False)
										self.addItem(newConn)
		#Connect Dummies node as they don't exist in the nodeList

	def calculateTotalTime(self):
		self.totalTime = 0
		for k, node in self.nodeList.items():
			if node.nodeContainer.isHostNode():
				self.totalTime += node.nodeContainer.getHostNode().computeTime

	def createNode(self, node):
		newNode = NodeFactory.Instance().buildNode(node, self.container)
		if newNode:
			self.addNode(newNode)
			if node.isSelected():
				newNode.setSelected(True)
			return newNode

	def createDummiesNode(self):
		if self.parent:
			newNode = PackedNetworkDummieNode(self.parent)
			newNode.editor = self.container
			self.dummiesNodeOutput = newNode
			self.nodeList[newNode.uid] = newNode
			self.addItem(newNode)
			newNode = PackedNetworkDummieNode(self.parent, True)
			newNode.editor = self.container
			self.dummiesNodeInput = newNode
			self.nodeList[newNode.uid + "-"] = newNode
			self.addItem(newNode)

	def buildGraph(self):
		for plugin in self.graphPlugin.itervalues():
			plugin.onGraphPreRebuild()

		nodes = self.hairSystem.getNodes()
		for node in nodes:
			if self.parent:
				if node.parent():
					if self.parent.uid == node.parent().uid:
						self.createNode(node)
			else:
				if not node.parent():
					self.createNode(node)
		self.createDummiesNode()
		self.rebuildConnection()

		for plugin in self.graphPlugin.itervalues():
			plugin.onGraphPostRebuild()
		#self.autoLayout()

	###
	### Event
	###

	def nodeSelected(self, node):
		if not node.isSelected():
			modifiers = QtWidgets.QApplication.keyboardModifiers()
			if modifiers == QtCore.Qt.ShiftModifier:
				node.setSelected(True)
				if node.nodeContainer.isHostNode():
					self.deferedCall = lambda: self.container.bridge.setSelection(node.nodeContainer.getHostNode().path, True)
			else:
				self.clearSelection()
				node.setSelected(True)
				if node.nodeContainer.isHostNode():
					self.deferedCall = lambda: self.container.bridge.setSelection(node.nodeContainer.getHostNode().path, False)

	def nodeMoving(self, node, delta):
		for selectedNode in self.selectedItems():
			if node != selectedNode:
				selectedNode.setPos(selectedNode.pos() + delta)
				selectedNode.update()
		node.setPos(node.pos() + delta)
		node.update()
		self.container.bridge.saveHairNetwork(self.hairSystem)

	def connectionSelected(self, connection):
		connection.setSelected(True)

	###
	### Utility
	###

	def autoLayout(self):
		if len(self.nodeList) > 0:
			autoLayout = AutoLayout(self.nodeList.values())
			autoLayout.doLayout()
			self.container.bridge.saveHairNetwork(self.hairSystem)

	###
	### Plugin Manager
	###

	def registerGraphPlugin(self, name, plugin):
		plugin.setGraph(self)
		self.graphPlugin[name] = plugin

	def deregisterGraphPlugin(self, name):
		del self.graphPlugin[name]


	###
	### Getter/Setter
	###

	def setHairSystem(self, hairSystem):
		if self.hairSystem:
			self.container.bridge.saveHairNetwork(self.hairSystem)
		self.hairSystem = hairSystem

	def setParent(self, parent):
		if parent is None:
			self.dummiesNodeOutput = None
			self.dummiesNodeInput = None
		self.parent = parent
		self.update()

	def getNewId(self):
		self.poolId += 1
		return self.poolId - 1


	def getNodeById(self, id):
		if id in self.nodeList:
			return self.nodeList[id]

	def addNode(self, node, Sig=True):
		node.graph = self
		self.nodeList[node.uid] = node
		self.addItem(node)

	def flushGraph(self):
		self.clear()
		self.nodeList = {}


	###
	### Virtual Override
	###

	def mousePressEvent(self, event):
		super(OxScene, self).mousePressEvent(event)


		#self.container.bridge.setSelection([], False)
		self.haveMoved = False
		self.beginClick = True

	def mouseMoveEvent(self, event):
		super(OxScene, self).mouseMoveEvent(event)
		self.haveMoved = True

	def mouseReleaseEvent(self, event):
		super(OxScene, self).mouseReleaseEvent(event)
		if self.deferedCall:
			self.deferedCall()

		if not event.isAccepted() and event.button() == QtCore.Qt.LeftButton:
			self.container.bridge.setSelection([], False)
		self.deferedCall = None
		self.ignoreNextCompute = False
		self.haveMoved = False
		self.beginClick = False

	#
	# Drag & Drop Handling
	#

	def dragEnterEvent(self, event):
		if event.mimeData().hasFormat('text/plain'):
			event.accept()
		else:
			event.ignore()

	def dragMoveEvent(self, event):
		if event.mimeData().hasFormat('text/plain'):
			event.accept()
		else:
			event.ignore()

	def dropEvent(self, event):
		if event.mimeData().text().startswith("newnode"):
			type = event.mimeData().text()[7:]
			node = self.container.bridge.createNode(type)
			if self.parent:
				self.container.bridge.setParent(node, self.parent.getHostNode())
			hairNode = self.hairSystem.insertHostNode(node)
			hairNode.setPos(event.scenePos().x(), event.scenePos().y())
		else:
			droppedNode = event.mimeData().text().split('\n')
			wraps = self.container.bridge.wrapNodes(droppedNode)
			i = 0
			for node in wraps:
				hairNode = self.hairSystem.insertHostNode(node)
				hairNode.setPos(event.scenePos().x(), event.scenePos().y() + (i * 100))
				i += 1
		self.update()

	#
	# Context Menu
	#
	def contextMenuEvent(self, event):
		item = self.itemAt(event.scenePos().toPoint(), QtGui.QTransform())
		if item:
			# Try run items context if it has one
			item.contextMenuEvent(event)
			return

		menu = NodeCreationMenu(self.container, event.scenePos())
		result = menu.exec_(event.screenPos())
		if result is not None:
			self.container.bridge.startUndo()
			if result.data()[0] == "node":
				node = self.container.bridge.createNode(result.data()[1])
				if self.parent:
					self.container.bridge.setParent(node, self.parent.getHostNode())
				hairNode = self.hairSystem.insertHostNode(node)
				hairNode.setPos(event.scenePos().x(), event.scenePos().y())
			else:
				from ...Container.HairSystemNode import *
				node = HairSystemNode("")
				node.type = "group"
				import uuid
				node.iud = str(uuid.uuid4())
				self.hairSystem.addNode(node)
			self.container.bridge.endUndo()
		self.update()


	###
	### Drawing
	###

	def drawBackground(self, painter, rect):
		gridSize = 25

		left = int(rect.left()) - (int(rect.left()) % gridSize)
		top = int(rect.top()) - (int(rect.top()) % gridSize)

		lines = []

		painter.setPen(
			QtGui.QPen(QtGui.QColor(20, 20, 20), 1, QtCore.Qt.DashLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))

		x = left
		while x < rect.right():
			if x != 0:
				lines.append(QtCore.QLineF(x, rect.top(), x, rect.bottom()))
			x += gridSize

		y = top
		while y < rect.bottom():
			if y != 0:
				lines.append(QtCore.QLineF(rect.left(), y, rect.right(), y))
			y += gridSize

		# painter.setRenderHint(QPainter.Antialiasing)
		painter.setRenderHint(QtGui.QPainter.SmoothPixmapTransform)
		painter.drawLines(lines);
		painter.setPen(
			QtGui.QPen(QtGui.QColor(20, 20, 20), 2.5, QtCore.Qt.DashDotLine, QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))
		painter.drawLine(QtCore.QLineF(0, rect.top(), 0, rect.bottom()));
		painter.drawLine(QtCore.QLineF(rect.left(), 0, rect.right(), 0))
