from Ornatrix.Graph.Bridge.BaseBridge import *
from Ornatrix.Graph.Container.HairSystem import HairSystem
from MayaBridgeLoader import MayaBridgeLoader


import maya.cmds as mc
import os

class MayaBridge(BaseBridge):

	_bridgeVersion = "1.0"

	def __init__(self):
		super(MayaBridge, self).__init__()

		mc.OxProfiler(enableNodeTimer=True)
		self.loadConfig()

		self.registerCallBack()

		self.previousHairSystemSel = None
		self.changed = False
		self.lock = False

	def __del__(self):
		super(MayaBridge, self).__del__()
		for job in self.job:
			mc.scriptJob(kill=job)
		self.job = []

	###
	### Bridge Initialisation
	###

	def registerCallBack(self):
		self.job = []
		self.job.append(mc.scriptJob(event=["DagObjectCreated", self.dataChanged]))
		self.job.append(mc.scriptJob(event=["SelectionChanged", self.dataChanged]))
		self.job.append(mc.scriptJob(event=["NewSceneOpened", self.clear]))
		self.job.append(mc.scriptJob(event=["SceneOpened", self.clear]))

		#self.job.append(mc.scriptJob(event=["idle", self.doChange]))

	def loadConfig(self):
		loader = MayaBridgeLoader(self)
		loader.load("/config.json")
	###
	### Slot
	###

	#Avoid Updating the UI multiple time by waiting for a maya Idle
	def doChange(self):
		if not self.lock:
			self.lock = True
			super(MayaBridge, self).dataChanged()
			self.changed = None
			self.lock = False

	def dataChanged(self):
		if not self.changed:
			self.changed = mc.scriptJob(event=["idle", self.doChange], runOnce=True)

	###
	### Bridge Getter
	###

	def getNodeConfig(self, type):
		if type not in self.hostNode:
			self.registerHostNode(type, {'canCreate': False})
		return self.hostNode[type]


	###
	### MAYA GATHERER
	###
	def getAllOxNode(self, nodeType=None):
		if nodeType is None:
			nodeType = self.getHostNode(True).keys()
		print nodeType
		nodes = mc.ls(type=nodeType)
		return self.wrapNodes(nodes)

	def getSelectedOxNodes(self, asShape=False):
		selectedNode = []
		nodes = mc.ls(sl=True, type=self.getHostNode().keys())
		for selected in nodes:
			if asShape:
				shapes = mc.listRelatives(selected, shapes=True)
				if shapes:
					for shape in shapes:
						selectedNode.append(self.wrapNode(shape))
				else:
					selectedNode.append(self.wrapNode(selected))
			else:
				selectedNode.append(self.wrapNode(selected))
		return selectedNode

	def getHairSystem(self):
		hairSystems = []
		sets = mc.ls(type="objectSet")
		for set in sets:
			if "hairSystem" in set:
				if set in self.preWrappedHairSystem:
					hairSystem = self.preWrappedHairSystem[set]
					hairSystem.unbindHostNode()
				else:
					hairSystem = HairSystem(set, set)
					if not mc.attributeQuery("data", node=set, exists=True):
						mc.addAttr(set, ln="data", dt="string")
					hairSystem.deserialize(mc.getAttr(set + ".data"))
					self.preWrappedHairSystem[set] = hairSystem
				for node in self.getHairSystemNode(hairSystem):
					hairSystem.bindHostNode(node)
				hairSystem.clean()
				hairSystems.append(hairSystem)
		return hairSystems

	def getHairSystemNode(self, hairSystem):
		return self.wrapNodes(mc.sets(hairSystem.path, q=True))

	# Todo : handle branching (return all highest node)
	def findHighestNode(self, node):
		upperNode = [node]
		while True:
			result = self.getNodeAbove(upperNode[0])
			if len(result) == 0:
				break
			upperNode = result
		return upperNode

	def findLowestNode(self, node, stopIfParentDifferent = False):
		lowerNode = [node]
		while True:
			result = self.getNodeUnder(lowerNode[0])
			if len(result) == 0:
				break
			if stopIfParentDifferent and node.getParent() != result[0].getParent():
				break
			lowerNode = result
		return lowerNode

	def isNodeAbove(self, target, start):
		if start.path == target.path:
			return True
		results = self.getNodeAbove(start)
		for result in results:
			return self.isNodeAbove(target, result)
		return False

	# Return a list of all node using this node as a source
	def getNodeAbove(self, sourceNode, id=0):
		output = self.getNodeConfig(sourceNode.type).getOutput(id)
		if not output:
			return []

		if mc.attributeQuery(output['attr'][0], node=sourceNode.path, exists=True):
			conn = mc.listConnections(sourceNode.path + "." + output['attr'][0], sh=True, d=True, source=False)
			if conn:
				return self.wrapNodes(set(conn))
		return []

	# Find the node under the destinationNode, looking at inputMesh/inputHair for compatibility reason
	def getNodeUnder(self, destNode, id=0):
		input = self.getNodeConfig(destNode.type).getInput(id)
		if not input:
			return []
		wrappedNodes = []

		if mc.attributeQuery(input['attr'][0], node=destNode.path, exists=True):
			sourceNodes = mc.listConnections(destNode.path + "." + input['attr'][0], sh=True, d=False, source=True,
			                                 p=True)
			if sourceNodes:
				for source in sourceNodes:
					s = source.split('.')
					wrappedNode = self.wrapNode(s[0])
					# Reverse search the connection pin Id between the dest and the source
					x = 0
					for output in self.getNodeConfig(wrappedNode.type).getOutputs():
						if s[1] in output['attr']:
							wrappedNode.addOutputConnection(x, destNode, id)
						x+=1
					wrappedNodes.append(wrappedNode)
				return wrappedNodes
		return []

	def getNetwork(self, nodes, result = []):
		for n in nodes:
			res = set([])
			res |= set(self.getNodeUnder(n))
			res |= set(self.getNodeAbove(n))
			for r in res:
				if r not in result:
					result.append(r)
					self.getNetwork([r], result)
		return result

	def nodeExist(self, node):
		return mc.objExists(node)

	def getParent(self, node):
		if mc.attributeQuery("parent", node=node.path, exists=True):
			parent = mc.listConnections(node.path + ".parent", s=True, p=False, d=False, c=False)
			if parent:
				return self.wrapNode(parent[0])
		return None

	def getChilds(self, node):
		if mc.attributeQuery("child", node=node.path, exists=True):
			childs = mc.listConnections(node.path + ".child", s=False, p=False, d=True, c=False)
			return self.wrapNodes(childs)
		return []


	###
	### Maya Modifier
	###

	def startUndo(self, ignore=False):
		if ignore:
			mc.undoInfo(stateWithoutFlush=False)
		else:
			mc.undoInfo(openChunk=True)

	def endUndo(self, ignore=False):
		if ignore:
			mc.undoInfo(stateWithoutFlush=True)
		else:
			mc.undoInfo(closeChunk=True)

	### Hair System

	def createHairSystem(self):
		set = mc.sets(name="hairSystem", empty=True)
		if not mc.attributeQuery("data", node=set, exists=True):
			mc.addAttr(set, ln="data", dt="string")
		return self.getHairSystem()[0]

	def saveHairNetwork(self, hairNetwork):
		mc.setAttr(hairNetwork.path + ".data", hairNetwork.serialize(), type="string")

	def insertNodeInHairSystem(self, hairSystem, node):
		if mc.nodeType(node.path) == "transform":
			node = self.wrapNode(mc.listRelatives(node.path, shapes=True)[0])
		if node:
			mc.sets(node.path, e=True, addElement=hairSystem.path)

	### General
	def createNode(self, type):
		newNode = mc.createNode(type)
		if not mc.attributeQuery("distributionMesh", node=newNode, exists=True):
			mc.addAttr(newNode, ln="distributionMesh", dt="mesh")
		if not mc.attributeQuery("distributionMeshMatrix", node=newNode, exists=True):
			mc.addAttr(newNode, ln="distributionMeshMatrix", dt="matrix")
		return self.wrapNode(newNode)

	def copyNode(self, node):
		return self.wrapNode(mc.duplicate(node.path)[0])

	def connectNode(self, source, destination, inputId=0, outputId=0):
		sourceConfig = self.getNodeConfig(source.type)
		destConfig = self.getNodeConfig(destination.type)



		outputs = sourceConfig.getOutput(outputId)
		inputs = destConfig.getInput(inputId)

		if not outputs:
			mc.warning("Ox Connection Failed : Node " + source.path + " missing output " + str(outputId))
			return
		if not inputs:
			mc.warning("Ox Connection Failed : Node " + destination.path + " missing input " + str(inputId))
			return


		for key in range(0, len(inputs['attr'])):
			if mc.attributeQuery(inputs['attr'][key], node=destination.path, exists=True):
				if mc.attributeQuery(outputs['attr'][key], node=source.path, exists=True):
					if mc.attributeQuery(inputs['attr'][key], node=destination.path, multi=True):
						start_index = 0
						while start_index < 10000000:
							if len(mc.connectionInfo('{}.{}[{}]'.format(destination.path, inputs['attr'][key], str(start_index)), sfd=True) or []) == 0:
								break
							start_index += 1
						mc.connectAttr(source.path + "." + outputs['attr'][key],
						               destination.path + "." + inputs['attr'][key] + "[" + str(start_index) + "]", force=True)
					else:
						mc.connectAttr(source.path + "." + outputs['attr'][key],
						               destination.path + "." + inputs['attr'][key], force=True)

	# Disconnect a source node from it destination
	def disconnectNode(self, source, destination, inputId=0, outputId=0):
		sourceConfig = self.getNodeConfig(source.type)
		destConfig = self.getNodeConfig(destination.type)

		outputs = sourceConfig.getOutput(outputId)
		inputs = destConfig.getInput(inputId)

		if not outputs:
			mc.warning("Ox Disconnection Failed : Node " + source.path + " missing output " + str(outputId))
			return


		dests = mc.listConnections(source.path, s=False, p=True, d=True, c=True)
		for i in range(0, len(dests), 2):
			name = dests[i + 1].split('.')[0]
			output = dests[i].split('.')[1]
			if name == destination.path and output in outputs['attr']:
				mc.disconnectAttr(dests[i], dests[i + 1])

	def setParent(self, node, parent):
		if parent:
			if not mc.attributeQuery("parent", node=node.path, exists=True):
				mc.addAttr(node.path, ln="parent", at="message")

			if not mc.attributeQuery("child", node=parent.path, exists=True):
				mc.addAttr(parent.path, ln="child", at="message")

			mc.connectAttr(parent.path + ".child", node.path + ".parent")
		else:
			if mc.attributeQuery("parent", node=node.path, exists=True):
				dests = mc.listConnections(node.path + ".parent", s=True, p=True, d=False, c=True)
				print dests
				for i in range(0, len(dests), 2):
					mc.disconnectAttr(dests[i+1], dests[i])

	def deleteNode(self, node):
		super(MayaBridge, self).deleteNode(node)
		mc.delete(node.path)

	def renameNode(self, node, newName):
		mc.rename(node.path, newName)
		self.dataChanged()

	def toggleNode(self, node):
		state = not mc.getAttr(node.path + ".nodeState")
		mc.setAttr(node.path + ".nodeState", state)

		self.dataChanged()

	def setSelection(self, name, add=False):
		mc.select(name, add=add)

	###
	### CONVERTER
	###
	def wrapNode(self, path):
		if len(path) == 0:
			return None
		name = path.split('|')[-1]
		selecteds = mc.ls(sl=True)
		type = mc.nodeType(path)
		typeApi = mc.nodeType(path, i=True)
		uuid = mc.ls(path, uuid=True)[0]

		nodeConfig = self.getNodeConfig(type)

		if uuid in self.preWrappedNode:
			nodeInfo = self.preWrappedNode[uuid]
		else:
			nodeInfo = HostNode(name, path)

			self.job.append(mc.scriptJob(attributeDeleted=(path+".message" , self.dataChanged), runOnce=True))
			try:
				if nodeConfig.getOutput(0):
					self.job.append(mc.scriptJob(attributeChange=(path + "." + nodeConfig.getOutput(0)['attr'][0], self.dataChanged)))
			except KeyError:
				pass


			nodeInfo.config = nodeConfig
			nodeInfo.icon = nodeConfig.getIcon()
			nodeInfo.container = self.container
			self.preWrappedNode[uuid] = nodeInfo
			nodeInfo.uid = uuid

			# Looking for compatible Input
			nodeInfo.type = type

			for input in nodeConfig.getInputs():
				nodeInfo.addInput(input['attr'][0], input['acceptedType'])
			for output in nodeConfig.getOutputs():
				print "adding output for node", type
				nodeInfo.addOutput(output['attr'][0], output['type'])

		try:
			if nodeConfig.getOutput(0):
				nodeInfo.computeTime = mc.OxProfiler(nodeComputeTime=path + "." + nodeConfig.getOutput(0)['attr'][0])
		except KeyError:
			nodeInfo.computeTime = mc.OxProfiler(nodeComputeTime=path)


		nodeInfo.typeApi = typeApi
		nodeInfo.selected = path in selecteds
		nodeInfo.enabled = True if mc.getAttr(path + '.nodeState') == 0 else False
		return nodeInfo


	def wrapNodes(self, paths):
		wrappedNodes = []
		if paths:
			for path in paths:
				wrappedNodes.append(self.wrapNode(path))
		return wrappedNodes

	def saveGroom(self):
		mc.OxSaveGroom()

	###
	### Deprecated // will be deleted as the graph get rewriten
	###

	# Return the hairSystem currently selected or last one if no Ox node are selected
