import maya.cmds as cmds
import maya.OpenMaya as om


def IsTransferrableStackShapeAttribute(attrName):
	return attrName.startswith('ai')


def CreateAttributeCopy(targetNode, attributeLongName, attributeType):
	if not cmds.attributeQuery(attributeLongName, node=targetNode, exists = True ):
		# print( 'Creating attribute {}.{} of type {}'.format( targetNode, attributeLongName, attributeType ) )
		if attributeType == 'string' or attributeType == 'float3':
			cmds.addAttr(targetNode, longName=attributeLongName, dataType=attributeType, hidden=True)
		elif attributeType == 'enum':
			cmds.addAttr(targetNode, longName=attributeLongName, attributeType=attributeType, hidden=True, minValue=0, maxValue=1000)
		else:
			cmds.addAttr(targetNode, longName=attributeLongName, attributeType=attributeType, hidden=True)


def GetTransferrableStackShapeAttributes(sourceShape):
	# print( 'GetTransferrableStackShapeAttributes start: {}'.format( sourceShape ) )
	attributesList = [x for x in cmds.listAttr(sourceShape) if IsTransferrableStackShapeAttribute(x)]
	result = {}
	for attrName in attributesList:
		type = None
		value = None
		try:
			type = cmds.getAttr( sourceShape + '.' + attrName, type = True )
			value = cmds.getAttr( sourceShape + '.' + attrName )
		except:
			pass
		if value is not None:
			result[attrName] = {
				'longName': cmds.attributeQuery(attrName, node=sourceShape, longName = True ),
				'type': type,
				'value': value }
	# print( 'GetTransferrableStackShapeAttributes end: {}, result: {}'.format( sourceShape, result ) )
	return result


def SetTransferrableStackShapeAttributes(targetShape, attributes):
	# print( 'SetTransferrableStackShapeAttributes start: {}'.format( targetShape ) )
	for attrName in attributes:
		attribute = attributes[attrName]
		if attribute['type'] in ['bool', 'byte', 'long', 'float', 'string', 'enum', 'float3']:
			CreateAttributeCopy(targetShape, attribute['longName'], attribute['type'])
			if attribute['value'] is not None:
				# print( 'Setting attribute {}.{} of type {} to value {}'.format( targetShape, attribute['longName'], attribute['type'], attribute['value'] ) )
				if attribute['type'] in ['string']:
					# These require specifying the type explicitly
					cmds.setAttr( targetShape + '.' + attrName, attribute['value'], type = attribute['type'] )
				elif attribute['type'] == 'float3':
					# For some reason float3 attributes are placed in a list
					value = attribute['value'][0]
					# And the float3 tuple must be unpacked with *
					cmds.setAttr( targetShape + '.' + attrName, type = attribute['type'], *value )
				else:
					# The rest require not specifying the type
					cmds.setAttr( targetShape + '.' + attrName, attribute['value'] )
#		else:
#			cmds.OxLog( 'SetTransferrableStackShapeAttributes: skipped attribute {}, unsupported type {}'.format( attrName, attribute['type'] ) )

'''
Register/Unregister VRay pre/post render callbacks
'''

def OxVRayPreRender(data):
	cmds.OxSetIsRendering( True )

def OxVRayPostRender(data):
	cmds.OxSetIsRendering( False )

OxVRayPreRenderCb = 0
OxVRayPostRenderCb = 0

def OxRemoveVRayUserCallbacks():
	global OxVRayPreRenderCb;
	global OxVRayPostRenderCb;
	if OxVRayPreRenderCb != 0:
		om.MUserEventMessage.removeCallback(OxVRayPreRenderCb)
		OxVRayPreRenderCb = 0
	if OxVRayPostRenderCb != 0:
		om.MUserEventMessage.removeCallback(OxVRayPostRenderCb)
		OxVRayPostRenderCb = 0

def OxCheckVRayUserCallbacks():
	global OxVRayPreRenderCb;
	global OxVRayPostRenderCb;

	renderer = cmds.getAttr('defaultRenderGlobals.currentRenderer')
	if renderer == "vray":
		if OxVRayPreRenderCb == 0:
			OxVRayPreRenderCb = om.MUserEventMessage.addUserEventCallback('VRayPreRender', OxVRayPreRender)
		if OxVRayPostRenderCb == 0:
			OxVRayPostRenderCb = om.MUserEventMessage.addUserEventCallback('VRayPostRender', OxVRayPostRender)
	else:
		OxRemoveVRayUserCallbacks()
